"""Module to convert yaml tags in yaml file to python dict with  cliconfig tags."""

from typing import Any, Dict, Optional, Tuple

import yaml


class TaggedNode:
    """Node class for tagged tree (with yaml tags)."""

    def __init__(self, value: Any, tag: str, *, is_config: bool) -> None:
        self.tag = tag
        self.value = value
        self.is_config = is_config


def tagged_constructor(
    loader: yaml.SafeLoader, tag_suffix: str, node: yaml.Node
) -> Any:
    """Build a tagged tree node from yaml node."""
    if isinstance(node, yaml.ScalarNode):
        return TaggedNode(loader.construct_scalar(node), tag_suffix, is_config=False)
    if isinstance(node, yaml.SequenceNode):
        return TaggedNode(loader.construct_sequence(node), tag_suffix, is_config=False)
    return TaggedNode(
        loader.construct_mapping(node), tag_suffix, is_config=True  # type: ignore
    )


def get_yaml_loader() -> Any:
    """Return a yaml loader to parse tags and build tagged tree."""
    loader = yaml.SafeLoader
    loader.add_multi_constructor("", tagged_constructor)  # type: ignore
    return loader


def insert_tags(tagged_tree: Any) -> Tuple[Any, Optional[str]]:
    """Build dict with cliconfig tag from tagged tree (with yaml tags)."""
    if isinstance(tagged_tree, TaggedNode) and tagged_tree.is_config:
        tag = tagged_tree.tag
        if tag:
            tag = tag.replace("!", "")
        out_dict = build_tree_from_dict(tagged_tree.value)
        return out_dict, tag
    if isinstance(tagged_tree, dict):
        out_dict = build_tree_from_dict(tagged_tree)
        return out_dict, None
    if isinstance(tagged_tree, TaggedNode):
        tag = tagged_tree.tag
        if tag:
            tag = tag.replace("!", "")
        if isinstance(tagged_tree.value, str):
            value = yaml.safe_load(tagged_tree.value)
        else:
            value = tagged_tree.value
        return value, tag
    return tagged_tree, None


def build_tree_from_dict(in_dict: Dict) -> Dict[str, Any]:
    """Build a dict with cliconfig tag from a dict of tagged trees."""
    out_dict: Dict[str, Any] = {}
    for key, value in in_dict.items():
        tree, tag = insert_tags(value)
        if tag:
            subtag = tag.replace("!", "")
            key = f"{key}@{subtag}"
        out_dict[key] = tree
    return out_dict



"""Built-in processing classes.

Built-in classes to apply pre-merge, post-merge, pre-save and post-load modifications
to dict with processing routines :mod:`.process_routines`.

They are the default processing used by the config routines :func:`.load_config`
and :func:`.make_config`.
"""
from typing import Any, Dict, List, Set

from cliconfig.base import Config
from cliconfig.process_routines import (
    merge_flat_paths_processing,
    merge_flat_processing,
)
from cliconfig.processing._type_parser import _isinstance, _parse_type
from cliconfig.processing.base import Processing
from cliconfig.tag_routines import clean_all_tags, clean_tag, dict_clean_tags, is_tag_in


class ProcessMerge(Processing):
    """Merge dicts just in time with ``@merge_after/_before/_add`` tags.

    Tag your key with ``@merge_after``, ``@merge_before`` or ``@merge_add``
    to load the config corresponding to the value (which is a yaml path) and
    merge it just before or after the current config. The merged dicts will be
    processed with pre-merge but not post-merge to ensure that merged
    configurations are recursively processed with pre-merge before applying a
    post-merge processing. It allows the merged configs to make references to
    each other (typically for copy) even without containing the merge tags
    itself.

    * '@merge_add' merges the dict corresponding to the path by allowing ONLY new keys
      It is a security check when you want to add a dict completely new,
      the typical usage for a default config splitted in several files.
    * '@merge_after' merge the dict corresponding to the path on the current dict
    * '@merge_before' merge the current dict on the dict corresponding to the path

    The processing is a pre-merge processing only and occurs before
    almost all of the other processings.
    Pre-merge order: -20.0

    Examples
    --------
    .. code-block:: yaml

        # config1.yaml
        a:
          b: 1
          b_path@merge_after: dict2.yaml

        # config2.yaml
        a.b: 2
        c_path@merge_add: config3.yaml

        # config3.yaml
        c: 3

    Before merging, the config1 is interpreted as the dict:

    ::

        {'a': {'b': 2, 'b_path': 'config2.yaml'}, 'c_path': 'config3.yaml', 'c': 3}

    If you replace '@merge_after' by '@merge_before', it will be:

    ::

        {'a': {'b': 1, 'b_path': 'config2.yaml'}, 'c_path': 'config3.yaml', 'c': 3}

    Finally, if you replace ``@merge_after`` by ``@merge_add``, it will raises an
    error because the key ``a.b`` already exists in the dict.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = -20.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "merge_after"):
                if not isinstance(val, str) or not val.endswith(".yaml"):
                    raise ValueError(
                        "Key with '@merge_after' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, "merge_after")] = val
                # Merge + process the dicts
                # NOTE: we allow new keys with security because the merge
                # following this pre-merge will avoid the creation of
                # new keys if needed.
                flat_config = merge_flat_paths_processing(
                    flat_config,
                    val,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    postprocess=False,
                )

            elif is_tag_in(flat_key, "merge_before"):
                if not isinstance(val, str) or not val.endswith(".yaml"):
                    raise ValueError(
                        "Key with '@merge_before' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, "merge_before")] = val
                # Merge + process the dicts
                flat_config = merge_flat_paths_processing(
                    val,
                    flat_config,
                    allow_new_keys=True,
                    preprocess_second=False,  # Already processed
                    postprocess=False,
                )

            elif is_tag_in(flat_key, "merge_add"):
                if not isinstance(val, str) or not val.endswith(".yaml"):
                    raise ValueError(
                        "Key with '@merge_add' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, "merge_add")] = val
                # Pre-merge process the dict with the process list of
                # the current config
                flat_config_to_merge = merge_flat_paths_processing(
                    Config({}, []),
                    val,
                    additional_process=flat_config.process_list,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    postprocess=False,
                )
                clean_dict, _ = dict_clean_tags(flat_config.dict)
                clean_dict_to_merge, _ = dict_clean_tags(flat_config_to_merge.dict)
                for key in clean_dict_to_merge:
                    if clean_all_tags(key) in clean_dict:
                        raise ValueError(
                            f"@merge_add doest not allow to add already "
                            f"existing keys but key '{key}' is found in both "
                            "dicts. Use @merge_after or @merge_before if you "
                            "want to merge this key, or check your key names."
                        )
                # Merge the dicts (order is not important by construction)
                # NOTE: we delete the process list of the current config
                # to speed up the process by avoiding redundant processing
                flat_config = merge_flat_processing(
                    Config(flat_config.dict, []),
                    flat_config_to_merge,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                    preprocess_second=False,  # Already processed
                    postprocess=False,
                )
        return flat_config


class ProcessCopy(Processing):
    """Copy a value with ``@copy`` tag. The copy is protected from direct updates.

    Tag your key with ``@copy`` and with value the name of the flat key to copy.
    The pre-merge processing removes the tag. The post-merge processing
    sets the value (if the copied key exists). The end-build processing checks
    that the copied key exists or was already copied once. The pre-save processing
    restore the tag and the key to copy to keep the information on future loads.
    The post-merge and the end-build processings occurs after most processings to
    allow the user to modify or add the copied key before the copy.
    Pre-merge order: 0.0
    Post-merge order: 10.0
    End-build order: 10.0
    Pre-save order: 0.0

    Examples
    --------
    .. code-block:: yaml

        # config.yaml
        a:
          b: 1
          c@copy: a.b

    Before merging, the config is interpreted as the dict

    .. code-block:: python

        {'a': {'b': 1, 'c': 1}}

    Note
    ----

        * The copy key is protected against any modification and will raise an error
          if you try to modify it but will be updated if the copied key is updated.
        * If the key to copy does not exist in the config on post-merge, no error
          is raised to let the user the possibility to add the key later via merge.
          However, if the key still does not exist at the end of the build
          (and the key was never copied), an error is raised.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 0.0
        self.postmerge_order = 10.0
        self.endbuild_order = 10.0
        self.presave_order = 0.0
        self.keys_to_copy: Dict[str, str] = {}
        self.copied_keys: Set[str] = set()
        self.current_value: Dict[str, Any] = {}

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "copy"):
                if not isinstance(val, str):
                    raise ValueError(
                        "Key with '@copy' tag must be associated "
                        "to a string corresponding to a flat key. "
                        f"The problem occurs at key: {flat_key} with value: {val}"
                    )
                clean_key = clean_all_tags(flat_key)
                if (
                    clean_key in self.keys_to_copy
                    and self.keys_to_copy[clean_key] != val
                ):
                    raise ValueError(
                        "Key with '@copy' has change its value to copy. Found key: "
                        f"{flat_key} with value: {val}, previous value to copy: "
                        f"{self.keys_to_copy[clean_key]}"
                    )
                # Store the key to copy and value
                self.keys_to_copy[clean_key] = val
                self.current_value[clean_key] = val
                # Remove the tag and update the dict
                flat_config.dict[clean_tag(flat_key, "copy")] = val
                del flat_config.dict[flat_key]
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        for key, val in self.keys_to_copy.items():
            # NOTE: Do not raise an error if the key to copy does not exist
            # yet because it can be added later in a future merge
            if key in flat_config.dict and val in flat_config.dict:
                if flat_config.dict[key] != self.current_value[key]:
                    # The key has been modified
                    raise ValueError(
                        "Found attempt to modify a key with '@copy' tag. The key "
                        f"is protected against direct updates. Found key: {key} of "
                        f"value {flat_config.dict[key]} that copy {val} of value "
                        f"{flat_config.dict[val]}"
                    )
                # Copy the value
                flat_config.dict[key] = flat_config.dict[val]
                self.copied_keys.add(key)
                # Update the current value
                self.current_value[key] = flat_config.dict[val]
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing."""
        for key, val in self.keys_to_copy.items():
            if key in flat_config.dict and key not in self.copied_keys:
                if val in flat_config.dict:
                    # Copy the value
                    flat_config.dict[key] = flat_config.dict[val]
                    self.copied_keys.add(key)
                else:
                    raise ValueError(
                        "A key with '@copy' tag has been found but the key to copy "
                        "does not exist at the end of the build and it has been "
                        f"never copied. Found key: {key} that would copy the "
                        f"key: {val}."
                    )
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the key to copy to keep the information
        # on further loading
        keys = list(flat_config.dict.keys())
        for key in keys:
            clean_key = clean_all_tags(key)
            if clean_key in self.keys_to_copy:
                new_key = key + "@copy"
                del flat_config.dict[key]
                flat_config.dict[new_key] = self.keys_to_copy[clean_key]
        return flat_config


class ProcessTyping(Processing):
    """Force a type with ``@type:<mytype>`` tag. The type is then forced forever.

    Allow basic types (none, any, bool, int, float, str, list, dict), nested lists,
    nested dicts, unions (with Union or the '|' symbol) and Optional.
    The type description is lowercased and spaces are removed.

    For instance: ``@type:None|List[Dict[str, int|float]]`` is valid and force
    the type to be None or a list containing dicts with str keys and int or float
    values.

    The processing stores the type in pre-merge and check alls forced types on
    end-build. It restore the tag in pre-save to keep the information on
    future loads. The end-build processing occurs after almost all processings.
    Pre-merge order: 0.0
    End-build order: 20.0
    Pre-save order: 0.0

    Note
    ----
        The type is not checked on pre-merge or ost-merge to allow the parameter
        to be updated (by a copy or a merge for instance). The goal of this
        processing is to ensure the type at the end of the build.

    Examples
    --------
    ::

        in_dict = {"param@type:None|List[int|float]": None}
        dict1 = {param: [0, 1, 2.0]}  # no error
        dict2 = {param: [0, 1, 2.0, 'a']}  # error

    Merging configs with dictionaries ``in_dict`` and ``dict1`` raises no
    error and ``param`` is forced to be None or a list of int or float forever.
    Merging config with ``in_dict`` and ``dict3`` raises an error on post-merge
    due to the 'a' value (which is a string).

    Note that removing "None|" in the type description of ``param`` still
    doesn't raise an error in the first case because the type checking is
    evaluated after the merge with ``dict2``.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 0.0
        self.endbuild_order = 20.0
        self.presave_order = 0.0
        self.forced_types: Dict[str, tuple] = {}
        self.type_desc: Dict[str, str] = {}  # For error messages

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            end_key = flat_key.split(".")[-1]
            if "@type:" in end_key:
                # Get the type description
                trail = end_key.split("@type:")[-1]
                type_desc = trail.split("@")[0]  # (in case of multiple tags)
                expected_type = tuple(_parse_type(type_desc))
                clean_key = clean_all_tags(flat_key)
                if clean_key in self.forced_types and set(
                    self.forced_types[clean_key]
                ) != set(expected_type):
                    raise ValueError(
                        f"Find the tag '@type:{type_desc}' on a key that has already "
                        "been associated to an other type: "
                        f"{self.type_desc[clean_key]}. "
                        f"Find problem at key: {flat_key}"
                    )
                # Remove the tag
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, f"type:{type_desc}")] = val
                # Store the forced type
                self.forced_types[clean_key] = expected_type
                self.type_desc[clean_key] = type_desc
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing."""
        for key, expected_type in self.forced_types.items():
            if key in flat_config.dict and not _isinstance(
                flat_config.dict[key], expected_type
            ):
                type_desc = self.type_desc[key]
                raise ValueError(
                    f"Key previously tagged with '@type:{type_desc}' must be "
                    f"associated to a value of type {type_desc}. Find the "
                    f"value: {flat_config.dict[key]} of type "
                    f"{type(flat_config.dict[key])} at key: {key}"
                )
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the type to keep the information
        # on further loading
        keys = list(flat_config.dict.keys())
        for key in keys:
            clean_key = clean_all_tags(key)
            if clean_key in self.type_desc:
                new_key = key + f"@type:{self.type_desc[clean_key]}"
                flat_config.dict[new_key] = flat_config.dict[key]
                del flat_config.dict[key]
        return flat_config


class ProcessSelect(Processing):
    """Select a sub-config with ``@select`` and delete the rest of its parent config.

    First look in pre-merge for a parameter tagged with ``@select`` containing a
    flat key corresponding to a sub-configurations to keep. The parent configuration
    is then deleted on post-merge, except the selected sub-configuration
    and eventually the tagged parameter (if it is in the same sub-configuration).
    It is also possible to select multiple keys of a same sub-configuration
    (meaning that the part before the last dot must be equal) by passing a
    list of flat keys.
    Pre-merge order: 0.0
    Post-merge order: 0.0

    Examples
    --------
    .. code-block:: yaml

        models:
            model_names@select: [models.model1, models.model3]
            model1:
                param1: 1
                param2: 2
            model2:
                param1: 3
                param2: 4
            model3:
                submodel:
                    param: 5
            model4:
                param: 6

    Result in deleting ``models.model2`` (``param1`` and ``param2``) and
    ``models.model4.param``, and keeping the rest.

    Warning
    -------

        For security reasons, this processing prevents from deleting
        the configuration at the root, which is the case when the
        selected key doesn't contain a dot. It raises an error in this case.
    """

    def __init__(self) -> None:
        super().__init__()
        self.keys_that_select: Set[str] = set()
        self.subconfigs_to_delete: Set[str] = set()
        self.keys_to_keep: Set[str] = set()

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, val in items:
            if is_tag_in(flat_key, "select"):
                # Remove the tag
                clean_key = clean_all_tags(flat_key)
                del flat_config.dict[flat_key]
                flat_config.dict[clean_tag(flat_key, "select")] = val
                self.keys_that_select.add(clean_key)
                if isinstance(val, str):
                    subconfig = ".".join(flat_key.split(".")[:-1])
                    keys_to_keep = [clean_key, val]
                elif isinstance(val, list):
                    subconfig = ".".join(val[0].split(".")[:-1])
                    for key in val[1:]:
                        subconfig2 = ".".join(key.split(".")[:-1])
                        if subconfig != subconfig2:
                            raise ValueError(
                                "The keys in the list of parameters tagged with "
                                "'@select' must be identical before the last dot "
                                f"(= on the same subconfig). Find: {subconfig} and "
                                f"{subconfig2} before the last dot."
                            )
                    keys_to_keep = [clean_key] + val
                else:
                    raise ValueError(
                        "The value of parameters tagged with '@select' must be a "
                        "string or a list of strings representing flat key(s). "
                    )
                subconfig = clean_all_tags(subconfig)
                if subconfig == "":
                    raise ValueError(
                        "Find attempt to delete the configuration at the root. You "
                        "must pass a flat key with a least one dot on parameter "
                        f"tagged with @select. Find key: {flat_key} with value: {val}"
                    )
                self.subconfigs_to_delete.add(subconfig)
                self.keys_to_keep.update(keys_to_keep)
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""

        def _is_in_subconfig(key: str, subconfig: str) -> bool:
            """Check if a key is in a subconfig with the exact name."""
            return key == subconfig or key.startswith(subconfig + ".")

        # Delete all keys on the subconfigs except the ones to keep
        for subconfig in self.subconfigs_to_delete:
            for key in list(flat_config.dict.keys()):
                if _is_in_subconfig(key, subconfig) and not any(
                    _is_in_subconfig(key, key_to_keep)
                    for key_to_keep in self.keys_to_keep
                ):
                    del flat_config.dict[key]
        return super().postmerge(flat_config)

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag with the type to keep the information
        # on further loading
        keys = list(flat_config.dict.keys())
        for key in keys:
            clean_key = clean_all_tags(key)
            if clean_key in self.keys_that_select:
                new_key = key + "@select"
                flat_config.dict[new_key] = flat_config.dict[key]
                del flat_config.dict[key]
        return flat_config


class ProcessDelete(Processing):
    """Delete the parameters tagged with ``@delete`` on pre-merge.

    This processing is useful to activate a processing without adding
    an additional parameter in the default configuration to avoid the error
    on merge with ``allow_new_keys=False``. This processing is applied very
    late on pre-merge to allow the others processing to be applied before
    deleting the parameters.
    Pre-merge order: 30.0

    Examples
    --------
    .. code-block:: yaml

        # main.yaml
        1@select@delete: configs.config1
        2@merge_add@delete: config1.yaml
        3@merge_add@delete: config2.yaml

        # config1.yaml
        configs.config1.param: 1
        configs.config1.param2: 2

        # config2.yaml
        configs.config2.param: 3
        configs.config2.param: 4

    Here we want to merge two config files and select one sub-config.
    We use the corresponding tags but we don't have a good name for the keys
    and instead of adding a new parameter in the default configuration with
    random names like "1", "2", "3", we use the ``@delete`` tag to delete the
    keys after the pre-merge processing.

    Warning
    -------

        The parameter is deleted on pre-merge. Therefore, if the parameter
        also exists on the other configuration during merge (without the tag),
        this parameter will be remain as it is. This processing is more used
        to delete parameter that is NOT present in the default configuration.
    """

    def __init__(self) -> None:
        super().__init__()
        # After all pre-merge processing
        self.premerge_order = 30.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        keys = list(flat_config.dict.keys())
        for key in keys:
            if is_tag_in(key, "delete"):
                del flat_config.dict[key]
        return flat_config


class ProcessNew(Processing):
    """Allow new sub-config/parameter absent from default config with tag ``@new``.

    The tagged sub-config/parameter and its value is stored during pre-merge and
    is deleted to avoid error on merge due to new key. It is then restored on
    post-merge. Finally, the tag is restored on pre-save for further loading.
    This processing is applied very late on pre-merge to allow
    the others processing to be applied before deleting the parameters.
    The post-merge processing is applied very early to allow the other processing
    to use this new parameter.
    Pre-merge order: 30.0
    Post-merge order: -20.0
    Pre-save order: 0.0

    Disclaimer: It is preferable to have an exhaustive default configuration instead
    of abusing this processing to improve the readability and to avoid typos errors
    in the name of the parameters or their sub-configs.

    Examples
    --------
    .. code-block:: yaml

        # default.yaml
        param1: 1

        # additional.yaml
        param2@new: 2
        subconfig@new.subsubconfig:
            param3: 3
            param4: 4

    Use default.yaml as default configuration and add additional.yaml as additional
    configuration via CLI results on the configuration containing param1, param2
    and the nested config containing param3 and param4.
    Without the ``@new`` tag, an error is raised because param2 is not present in
    the default configuration.

    Note
    ----

        * Tag a subconfig by adding ``@new`` at the end of the key containing
          the sub-config dict in your yaml file.
        * When a parameter is added with this processing, it is possible to modify it
          later via config merge without the tag because the parameter is then present
          in the current configuration.
        * If the tagged parameter or sub-config is already present in the current
          configuration, no error are raised and the value is still updated on
          post-merge. It may no have influence in practice.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 30.0
        self.postmerge_order = -20.0
        self.new_vals: Dict[str, Any] = {}
        self.new_vals_backup: Dict[str, Any] = {}

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        keys = list(flat_config.dict.keys())
        for key in keys:
            # NOTE: we don't use is_tag_in because we want to look
            # for tags in the sub-configs too.
            if "@new@" in key or "@new." in key or key.endswith("@new"):
                clean_key = clean_all_tags(key)
                self.new_vals[clean_key] = flat_config.dict[key]
                self.new_vals_backup[clean_key] = flat_config.dict[key]
                del flat_config.dict[key]
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        flat_config.dict.update(self.new_vals)
        # Reset the new values to avoid re-adding them later
        self.new_vals = {}
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        # Restore the tag @new to allow loading the config later by allowing
        # these new parameters.
        for key, value in self.new_vals_backup.items():
            if key in flat_config.dict:
                flat_config.dict[key + "@new"] = value
                del flat_config.dict[key]
        return flat_config


class ProcessCheckTags(Processing):
    """Raise an error if a tag is present in a key after pre-merging processes.

    This security processing is always applied after all pre-merge process and
    checks for '@' in the keys. It raises an error if one is found.
    """

    def __init__(self) -> None:
        super().__init__()
        # NOTE: this processing is a special meta-processing that must be
        # applied after all other pre-merge processing to ensure security.
        # That why it has a very high pre-merge order and it is not a
        # good idea to make pre-merge processing with higher order.
        self.premerge_order = 1000.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        _, tagged_keys = dict_clean_tags(flat_config.dict)
        if tagged_keys:
            keys_message = "\n".join(tagged_keys[:5])
            raise ValueError(
                "Keys with tags are encountered at the end of "
                "the pre-merge process. It is probably a mistake due to:\n"
                "- a typo in tag name\n"
                "- a missing processing in process_list\n"
                "- the use of an 'at' ('@') in a parameter name\n"
                "- the use of a custom processing that does not remove a tag\n\n"
                "The tagged keys encountered (5 first if more than 5) are:\n"
                f"{keys_message}"
            )
        return flat_config


class DefaultProcessings:
    """Default list of built-in processings.

    To add these processings to a Config instance, use:
    ::

        config.process_list += DefaultProcessings().list

    The current default processing list contains:
     * ProcessCheckTags: protect against '@' in keys at the end of pre-merge)
     * ProcessMerge (@merge_all, @merge_before, @merge_after): merge multiple
       files into one.
     * ProcessCopy (@copy): persistently copy a value from one key to an other
       and protect it
     * ProcessTyping (@type:X): force the type of parameter to any type X.
     * ProcessSelect (@select): select sub-config(s) to keep and delete the
       other sub-configs in the same parent config.
     * ProcessDelete (@delete): delete the parameter tagged with @delete on
       pre-merge.
    """

    def __init__(self) -> None:
        self.list: List[Processing] = [
            ProcessCheckTags(),
            ProcessMerge(),
            ProcessCopy(),
            ProcessTyping(),
            ProcessSelect(),
            ProcessDelete(),
            ProcessNew(),
        ]



"""Test built-in processing classes."""
import re

import pytest
import pytest_check as check

from cliconfig.base import Config
from cliconfig.dict_routines import flatten
from cliconfig.processing.builtin import (
    DefaultProcessings,
    ProcessCheckTags,
    ProcessCopy,
    ProcessDelete,
    ProcessMerge,
    ProcessNew,
    ProcessSelect,
    ProcessTyping,
)


def test_process_merge() -> None:
    """Test ProcessMerge."""
    # Test @merge_add
    processing = ProcessMerge()
    flat_dict = {
        "config1.param1": 0,
        "config1.param2": 1,
        "config2_path@merge_add": "tests/configs/merge/default2.yaml",
    }
    flat_config = Config(flat_dict, [processing])
    flat_config = processing.premerge(flat_config)
    expected_dict = {
        "config1.param1": 0,
        "config1.param2": 1,
        "config2.param1": 2,
        "config2.param2": 3,
        "config3.param1": 4,
        "config3.param2": 5,
        "config2_path": "tests/configs/merge/default2.yaml",
        "config3_path": "tests/configs/merge/default3.yaml",
    }
    check.equal(flat_config.dict, expected_dict)

    # Case of introducing a new key
    flat_dict = {
        "config1.param1": 0,
        "config1.param2": 1,
        "config3.param1": 4,
        "config2_path@merge_add": "tests/configs/merge/default2.yaml",
    }
    with pytest.raises(
        ValueError,
        match=(
            "@merge_add doest not allow to add "
            "already existing keys but key 'config3.param1'.*"
        ),
    ):
        processing.premerge(Config(flat_dict, [processing]))

    # Test @merge_before and @merge_after
    flat_dict = {
        "a.b": 1,
        "a.b_path@merge_after": "tests/configs/merge/additional2.yaml",
    }
    flat_config = Config(flat_dict, [processing])
    flat_config = processing.premerge(flat_config)
    expected_dict = {
        "a.b": 2,
        "a.b_path": "tests/configs/merge/additional2.yaml",
        "c_path": "tests/configs/merge/additional3.yaml",
        "c": 3,
    }
    check.equal(flat_config.dict, expected_dict)
    flat_dict = {
        "a.b": 1,
        "a.b_path@merge_before": "tests/configs/merge/additional2.yaml",
        "c": 3,
    }
    flat_config = Config(flat_dict, [processing])
    flat_config = processing.premerge(flat_config)
    expected_dict = {
        "a.b": 1,
        "a.b_path": "tests/configs/merge/additional2.yaml",
        "c_path": "tests/configs/merge/additional3.yaml",
        "c": 3,
    }
    check.equal(flat_config.dict, expected_dict)
    check.equal(flat_config.process_list, [processing])

    # Not valid path
    for tag in ["merge_before", "merge_after", "merge_add"]:
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Key with '@{tag}' tag must be associated "
                "to a string corresponding to a *yaml* file."
                f"The problem occurs at key: a@{tag}"
            ),
        ):
            processing.premerge(Config({f"a@{tag}": "no_yaml"}, [processing]))


def test_process_copy() -> None:
    """Test ProcessCopy."""
    processing = ProcessCopy()
    flat_dict = {
        "config1.param1": 1,
        "config2.param2@copy": "config1.param1",
    }
    flat_config = Config(flat_dict, [processing])
    flat_config = processing.premerge(flat_config)
    check.equal(
        flat_config.dict, {"config1.param1": 1, "config2.param2": "config1.param1"}
    )
    flat_config.dict["config1.param1"] = 2
    flat_config = processing.postmerge(flat_config)
    check.equal(flat_config.dict, {"config1.param1": 2, "config2.param2": 2})
    check.equal(processing.copied_keys, {"config2.param2"})
    flat_config = processing.presave(flat_config)
    check.equal(processing.current_value, {"config2.param2": 2})
    check.equal(
        flat_config.dict, {"config1.param1": 2, "config2.param2@copy": "config1.param1"}
    )
    check.equal(processing.keys_to_copy, {"config2.param2": "config1.param1"})
    check.equal(flat_config.process_list, [processing])
    # Reset copy processing
    processing.keys_to_copy = {}
    # Case of wrong key
    with pytest.raises(
        ValueError,
        match=(
            "Key with '@copy' tag must be associated "
            "to a string corresponding to a flat key. "
            "The problem occurs at key: a@copy with value: True"
        ),
    ):
        processing.premerge(Config({"a@copy": True}, [processing]))
    # Case of already existing @copy but associated to an other key
    processing.keys_to_copy = {"a": "b"}
    with pytest.raises(
        ValueError,
        match=(
            "Key with '@copy' has change its value to copy. Found key: a@copy@tag "
            "with value: c, previous value to copy: b"
        ),
    ):
        processing.premerge(Config({"a@copy@tag": "c"}, [processing]))
    # Case of non-existing key (on post-merge): do not raise error
    processing.keys_to_copy = {"a": "b"}
    processing.postmerge(Config({"a": "b"}, [processing]))
    # Case of non-existing key (on end-build): raise error
    with pytest.raises(
        ValueError,
        match=re.escape(
            "A key with '@copy' tag has been found but the key to copy does not "
            "exist at the end of the build and it has been never copied. Found key: "
            "a that would copy the key: b."
        ),
    ):
        processing.endbuild(Config({"a": "b"}, [processing]))
    # Copy if it appears on end-build
    config = processing.endbuild(Config({"a": "b", "b": 3}, [processing]))
    check.equal(config.dict["a"], 3)
    # Case overwriting a key
    processing.current_value = {"a": 2}
    with pytest.raises(
        ValueError,
        match=(
            "Found attempt to modify a key with '@copy' tag. The key is "
            "protected against direct updates. Found key: a of value c that copy "
            "b of value 1"
        ),
    ):
        processing.postmerge(Config({"a": "c", "b": 1}, [processing]))


def test_process_typing() -> None:
    """Test ProcessTyping."""
    processing = ProcessTyping()
    flat_dict = {
        "param1@type:int": 1,
        "param2@type:List[Optional[Dict[str, int|float]]]": [{"a": [0.0]}],
    }
    flat_config = Config(flat_dict, [processing])
    flat_config = processing.premerge(flat_config)
    flat_config.dict["param1"] = 3
    flat_config.dict["param2"] = {"a": None, "b": [{"c": 1}]}
    flat_config.dict = flatten(flat_config.dict)
    flat_config = processing.postmerge(flat_config)
    check.equal(
        flat_config.dict, {"param1": 3, "param2.a": None, "param2.b": [{"c": 1}]}
    )
    check.equal(
        flat_config.process_list[0].forced_types,  # type: ignore
        {
            "param1": (int,),
            "param2": (("list", ((type(None), ("dict", (str,), (int, float))),)),),
        },
    )
    check.equal(
        flat_config.process_list[0].type_desc,  # type: ignore
        {"param1": "int", "param2": "List[Optional[Dict[str, int|float]]]"},
    )
    flat_config = processing.endbuild(flat_config)  # no error
    flat_config = processing.presave(flat_config)
    check.equal(
        flat_config.dict,
        {"param1@type:int": 3, "param2.a": None, "param2.b": [{"c": 1}]},
    )
    processing.forced_types = {}  # Reset forced types
    processing.type_desc = {}  # Reset type description

    # Case of different type on pre-merge: do not raise error!
    processing.premerge(Config({"param@type:int": "str"}, [processing]))

    # Case of already existing type and othere in premerge
    processing.forced_types = {"param": (int,)}
    processing.type_desc = {"param": "int"}
    with pytest.raises(
        ValueError,
        match=(
            "Find the tag '@type:str' on a key that has already been associated "
            "to an other type: int. Find problem at key: param@type:str"
        ),
    ):
        processing.premerge(Config({"param@type:str": "str"}, [processing]))

    # Case of wrong type in post-merge: no error
    processing.forced_types = {"param": (int,)}
    processing.type_desc = {"param": "int"}
    processing.postmerge(Config({"param": "mystr"}, [processing]))
    # Case of wrong type in end-build: raise error
    with pytest.raises(
        ValueError,
        match=(
            "Key previously tagged with '@type:int' must be "
            "associated to a value of type int. Find the "
            "value: mystr of type <class 'str'> at key: param"
        ),
    ):
        processing.endbuild(Config({"param": "mystr"}, [processing]))


def test_process_select() -> None:
    """Test ProcessSelect."""
    processing = ProcessSelect()
    flat_dict = {
        "models.model_names@select": ["models.model1", "models.model3"],
        "models.model1.param1": 1,
        "models.model1.param2": 2,
        "models.model2.param1": 3,
        "models.model2.param2": 4,
        "models.model3.submodel.param": 5,
        "models.model4.param": 6,
    }
    expected_dict = {
        "models.model_names": ["models.model1", "models.model3"],
        "models.model1.param1": 1,
        "models.model1.param2": 2,
        "models.model3.submodel.param": 5,
    }
    config = Config(flat_dict, [])
    config = processing.premerge(config)
    config = processing.postmerge(config)
    check.equal(config.dict, expected_dict)
    config = processing.presave(config)
    check.is_in("models.model_names@select", config.dict)
    check.equal(
        config.dict["models.model_names@select"], ["models.model1", "models.model3"]
    )
    check.is_not_in("models.model_names", config.dict)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The keys in the list of parameters tagged with '@select' must be "
            "identical before the last dot (= on the same subconfig). Find: "
            "abab and dede before the last dot."
        ),
    ):
        processing.premerge(Config({"p@select": ["abab.cdcd", "dede.fgfg"]}, []))
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The value of parameters tagged with '@select' must be a string or a "
            "list of strings representing flat key(s). "
        ),
    ):
        processing.premerge(Config({"p@select": 0}, []))
    with pytest.raises(
        ValueError,
        match=(
            "Find attempt to delete the configuration at the root. You must pass a "
            "flat key with a least one dot on parameter tagged with @select. "
            "Find key: p@select with value: root"
        ),
    ):
        processing.premerge(Config({"p@select": "root"}, []))


def test_process_delete() -> None:
    """Test ProcessDelete."""
    processing = ProcessDelete()
    config = Config(
        {
            "@select@delete": "configs.config1",
            "1@merge_add@delete": "config1.yaml",
            "2@merge_add@delete": "config2.yaml",
        },
        [processing],
    )
    config = processing.premerge(config)
    check.equal(config.dict, {})


def test_process_new() -> None:
    """Test ProcessNew."""
    processing = ProcessNew()
    flat_dict = {
        "param1": 1,
        "config1": {
            "param2": 2,
            "subconfig@new": {"param3": 3, "param4": 4},
        },
        "config2@new@tag.subconfig.param3": 3,
        "config3.subconfig@new.param4": 4,
    }
    flat_dict = flatten(flat_dict)
    config = Config(flat_dict, [processing])
    check.equal(
        processing.premerge(config).dict,
        {
            "param1": 1,
            "config1.param2": 2,
        },
    )
    check.equal(
        processing.postmerge(config).dict,
        {
            "param1": 1,
            "config1.param2": 2,
            "config1.subconfig.param3": 3,
            "config1.subconfig.param4": 4,
            "config2.subconfig.param3": 3,
            "config3.subconfig.param4": 4,
        },
    )
    check.equal(
        processing.presave(config).dict,
        {
            "param1": 1,
            "config1.param2": 2,
            "config1.subconfig.param3@new": 3,
            "config1.subconfig.param4@new": 4,
            "config2.subconfig.param3@new": 3,
            "config3.subconfig.param4@new": 4,
        },
    )


def test_process_check_tags() -> None:
    """Test ProcessCheckTags."""
    processing = ProcessCheckTags()
    flat_dict = {
        "config.param1": 1,
        "param1": 2,
    }
    config = Config(flat_dict, [processing])
    check.equal(processing.premerge(config).dict, flat_dict)

    flat_dicts = [{"param1@tag": 1}, {"@foo": 2}, {"@": 3}]
    for flat_dict in flat_dicts:
        with pytest.raises(
            ValueError,
            match=(
                "Keys with tags are encountered at the end of "
                "the pre-merge process.*"
            ),
        ):
            processing.premerge(Config(flat_dict, [processing]))


def test_default_processings() -> None:
    """Test DefaultProcessings."""
    config = Config({}, DefaultProcessings().list)
    for proc in [
        ProcessCheckTags(),
        ProcessMerge(),
        ProcessCopy(),
        ProcessTyping(),
        ProcessDelete(),
        ProcessSelect(),
        ProcessNew(),
    ]:
        check.is_in(proc, config.process_list)



"""Manage Pylint output on workflow."""
import sys

from github_actions_utils.color import score_to_hex_color


def check_output() -> float:
    """Check output of Pylint.

    Raises
    ------
    ValueError
        If Pylint score is below SCORE_MIN.

    Returns
    -------
    score: float
        Score of Pylint.
    """
    args = sys.argv[1:]
    score = -10000.0  # A default value
    for arg in args:
        if arg.startswith("--score="):
            score = float(arg.split("=")[1])
    if score == -10000.0:
        raise ValueError("Please specify the score of Pylint using the flag --score=N.")
    if score < PYLINT_SCORE_MIN:
        raise ValueError(
            f"Pylint score {score} is lower than minimum ({PYLINT_SCORE_MIN})"
        )

    return score


def main() -> None:
    """Check score and print it."""
    score = check_output()
    # Print color to be used in GitHub Actions
    print(score_to_hex_color(score, PYLINT_SCORE_MIN, PYLINT_SCORE_MAX))


if __name__ == "__main__":
    # PYLINT_SCORE_MIN can be changed safely depending on your needs.
    PYLINT_SCORE_MIN = 8.5
    PYLINT_SCORE_MAX = 10.0
    main()



"""CLI Config: build your configuration from CLI by merging with processing.

Copyright © 2023  Valentin Goldité

    This program is free software: you can redistribute it and/or modify
    it under the terms of the MIT License.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    This project is free to use for COMMERCIAL USE, MODIFICATION,
    DISTRIBUTION and PRIVATE USE as long as the original license is
    include as well as this copy right notice.
"""
from cliconfig._version import __version__, __version_tuple__
from cliconfig.base import Config
from cliconfig.config_routines import load_config, make_config, save_config, show_config
from cliconfig.process_routines import (
    merge_flat_paths_processing,
    merge_flat_processing,
)
from cliconfig.processing.base import Processing
from cliconfig.processing.builtin import DefaultProcessings
from cliconfig.processing.create import (
    create_processing_keep_property,
    create_processing_value,
)

__all__ = [
    "__version__",
    "__version_tuple__",
    "Config",
    "DefaultProcessings",
    "Processing",
    "create_processing_keep_property",
    "create_processing_value",
    "make_config",
    "load_config",
    "merge_flat_paths_processing",
    "merge_flat_processing",
    "save_config",
    "show_config",
]



"""Base classes of Config object."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional

# Imports Processing for mypy type checking only
if TYPE_CHECKING:
    from cliconfig.processing.base import Processing


class Config:
    """Class for configuration.

    Config object contain the config dict and the processing list
    and no methods except ``__init__``, ``__repr__``, ``__eq__``,
    ``__getattribute__``, ``__setattr__`` and ``__delattr__``.
    The Config objects are mutable and not hashable.

    Parameters
    ----------
    config_dict : Dict[str, Any]
        The config dict.
    process_list : Optional[List[Processing]], optional
        The list of Processing objects. If None, an empty list is used.
        The default is None.
    """

    def __init__(
        self,
        config_dict: Dict[str, Any],
        process_list: Optional[List["Processing"]] = None,
    ) -> None:
        self.dict = config_dict
        self.process_list = process_list if process_list else []

    def __repr__(self) -> str:
        """Representation of Config object."""
        process_classes = [process.__class__.__name__ for process in self.process_list]
        return f"Config({self.dict}, {process_classes})"

    def __eq__(self, other: Any) -> bool:
        """Equality operator.

        Two Config objects are equal if their dicts are equal and their
        lists of Processing objects are equal (order doesn't matter).
        """
        if (
            isinstance(other, Config)
            and self.dict == other.dict
            and len(self.process_list) == len(other.process_list)
        ):
            equal = True
            for processing in self.process_list:
                equal = equal and processing in other.process_list
            for processing in other.process_list:
                equal = equal and processing in self.process_list
            return equal
        return False

    def __getattribute__(self, __name: str) -> Any:
        """Get attribute, sub-configuration or parameter.

        The dict should be nested (unflattened). If it is not the case,
        you can apply :func:`dict_routines.flatten` on ``config.dict``
        to unflatten it.
        """
        if __name in ["dict", "process_list"]:
            return super().__getattribute__(__name)
        if __name not in self.dict:
            keys = ", ".join(self.dict.keys())
            raise AttributeError(  # pylint: disable=raise-missing-from
                f"Config has no attribute '{__name}'. Available keys are: {keys}."
            )
        if isinstance(self.dict[__name], dict):
            # If the attribute is a dict, return a Config object
            # so that we can access the nested keys with multiple dots
            return Config(self.dict[__name], process_list=self.process_list)
        return self.dict[__name]

    def __setattr__(self, __name: str, value: Any) -> None:
        """Set attribute, sub-configuration or parameter.

        The dict should be nested (unflattened). If it is not the case,
        you can apply :func:`dict_routines.flatten` on ``config.dict``
        to unflatten it.
        """
        if __name in ["dict", "process_list"]:
            super().__setattr__(__name, value)
        else:
            self.dict[__name] = value

    def __delattr__(self, __name: str) -> None:
        """Delete attribute, sub-configuration or parameter.

        The dict should be nested (unflattened). If it is not the case,
        you can apply :func:`dict_routines.flatten` on ``config.dict``
        to unflatten it.
        """
        if __name in ["dict", "process_list"]:
            super().__delattr__(__name)
        else:
            del self.dict[__name]



"""Manage Pydocstyle output on workflow."""
import sys


def check_output() -> None:
    """Check output of Pydocstyle.

    Raises
    ------
    ValueError
        If Pydocstyle find errors.
    """
    args = sys.argv[1:]
    n_errors = -1  # A default value
    for arg in args:
        if arg.startswith("--n_errors="):
            n_errors = int(arg.split("=")[1])
    if n_errors == -1:
        raise ValueError(
            "Please specify the number of errors found by Pydocstyle "
            "using the flag --n_errors=N."
        )
    if n_errors > 0:
        raise ValueError(
            f"Pydocstyle found {n_errors} errors in python "
            "docstrings. Please fix them.",
        )


if __name__ == "__main__":
    check_output()



"""Functions to create new processing quickly."""
# pylint: disable=unused-argument
import re
from inspect import signature
from typing import Any, Callable, Dict, Optional, Set, Union

from cliconfig.base import Config
from cliconfig.processing.base import Processing
from cliconfig.tag_routines import clean_all_tags, clean_tag, is_tag_in


def _is_matched(key: str, tag_name: Optional[str], regex: Optional[str]) -> bool:
    """Check if key match the regex or contain the tag."""
    if tag_name is not None and regex is not None:
        raise ValueError("Either regex or tag_name must be defined but not both.")
    # Case defined with tag
    if tag_name is not None:
        return is_tag_in(key, tag_name)
    # Case defined with regex
    if regex is not None:
        param_name = key.split(".")[-1].split("@")[0]
        return re.match(regex, param_name) is not None
    raise ValueError("Either regex or tag_name must be defined.")


class _ProcessingValue(Processing):
    """Processing class for make_processing_value."""

    def __init__(
        self,
        func: Union[Callable[[Any], Any], Callable[[Any, Config], Any]],
        processing_type: str,
        *,
        regex: Optional[str],
        tag_name: Optional[str],
        order: float,
        persistent: bool,
    ) -> None:
        super().__init__()
        self.func = func
        self.processing_type = processing_type
        self.regex = regex
        self.tag_name = tag_name
        if processing_type == "premerge":
            self.premerge_order = order
        elif processing_type == "postmerge":
            self.postmerge_order = order
        elif processing_type == "endbuild":
            self.endbuild_order = order
        elif processing_type == "presave":
            self.presave_order = order
        elif processing_type == "postload":
            self.postload_order = order
        else:
            raise ValueError(
                f"Processing type '{processing_type}' not recognized. Must be"
                "one of 'premerge', 'postmerge', 'endbuild', 'presave' or"
                "'postload'."
            )
        self.persistent = persistent
        # matched_keys: store keys that match the regex or contain the tag
        self.matched_keys: Set[str] = set()

    def _apply_update(self, flat_config: Config) -> Config:
        """Apply value updates to matching parameters.

        Does not assume that the keys are clean (it is not always the case
        on pre-merge and post-load for instance).
        """
        items = list(flat_config.dict.items())
        for key, value in items:
            clean_key = clean_all_tags(key)
            if clean_key in self.matched_keys:
                # Remove clean_key from match_keys if not persistent
                if not self.persistent:
                    self.matched_keys.remove(clean_key)

                # Apply update

                value = flat_config.dict[key]
                try:
                    # Case one argument
                    if len(signature(self.func).parameters) == 1:
                        flat_config.dict[key] = self.func(value)  # type: ignore
                    # Case two arguments
                    else:
                        flat_config.dict[key] = self.func(  # type: ignore
                            value, flat_config
                        )
                except ValueError:
                    # Case no signature, trying to call with one argument
                    flat_config.dict[key] = self.func(value)  # type: ignore
        return flat_config

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if _is_matched(flat_key, self.tag_name, self.regex):
                # Store the key
                self.matched_keys.add(clean_all_tags(flat_key))
                # Remove the tag if any
                if self.tag_name:
                    del flat_config.dict[flat_key]
                    new_key = clean_tag(flat_key, self.tag_name)
                    flat_config.dict[new_key] = value

        if self.processing_type == "premerge":
            return self._apply_update(flat_config)
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        if self.processing_type == "postmerge":
            return self._apply_update(flat_config)
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing."""
        if self.processing_type == "endbuild":
            return self._apply_update(flat_config)
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        if self.processing_type == "presave":
            return self._apply_update(flat_config)
        return flat_config

    def postload(self, flat_config: Config) -> Config:
        """Post-load processing."""
        if self.processing_type == "postload":
            return self._apply_update(flat_config)
        return flat_config


class _ProcessingKeepProperty(Processing):
    """Processing class for make_processing_keep_property."""

    def __init__(
        self,
        func: Union[Callable[[Any], Any], Callable[[Any, Config], Any]],
        *,
        regex: Optional[str],
        tag_name: Optional[str],
        premerge_order: float,
        postmerge_order: float,
        endbuild_order: float,
    ) -> None:
        super().__init__()
        self.func = func
        self.regex = regex
        self.tag_name = tag_name
        self.premerge_order = premerge_order
        self.postmerge_order = postmerge_order
        self.endbuild_order = endbuild_order

        self.properties: Dict[str, Any] = {}

    def _eval_property(self, key: str, flat_config: Config) -> Any:
        """Evaluate property."""
        try:
            # Case one argument
            if len(signature(self.func).parameters) == 1:
                return self.func(flat_config.dict[key])  # type: ignore
            # Case two arguments
            return self.func(flat_config.dict[key], flat_config)  # type: ignore
        except ValueError:
            # Case no signature, trying to call with one argument
            return self.func(flat_config.dict[key])  # type: ignore

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if _is_matched(flat_key, self.tag_name, self.regex):
                clean_key = clean_all_tags(flat_key)
                if clean_key not in self.properties:
                    property_ = self._eval_property(flat_key, flat_config)
                    self.properties[clean_key] = property_
                if self.tag_name:
                    del flat_config.dict[flat_key]
                    new_key = clean_tag(flat_key, self.tag_name)
                    flat_config.dict[new_key] = value
        return flat_config

    def _check_properties(self, flat_config: Config, processing_timing: str) -> None:
        """Check if all properties are still the same."""
        for flat_key, expected_property in self.properties.items():
            if flat_key in flat_config.dict:
                property_ = self._eval_property(flat_key, flat_config)
                if property_ != expected_property:
                    raise ValueError(
                        f"Property of key {flat_key} has changed from "
                        f"{expected_property} to "
                        f"{property_} while it is protected by a keep-property "
                        f"processing (problem found on {processing_timing})."
                    )

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        self._check_properties(flat_config, "post-merge")
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing."""
        self._check_properties(flat_config, "end-build")
        return flat_config


def create_processing_value(
    func: Union[Callable[[Any], Any], Callable[[Any, Config], Any]],
    processing_type: str = "premerge",
    *,
    regex: Optional[str] = None,
    tag_name: Optional[str] = None,
    order: float = 0.0,
    persistent: bool = False,
) -> Processing:
    r"""Create a processing object that modifies a value in config using tag or regex.

    The processing is applied on pre-merge. It triggers when the key matches
    the tag or the regex. The function apply ``flat_dict[key] = func(flat_dict[key])``.
    You must only provide one of tag or regex. If tag is provided, the tag will be
    removed from the key during pre-merge.

    It also possible to pass the flat config as a second argument of the function
    ``func``. In this case, the function apply
    ``flat_dict[key] = func(flat_dict[key], flat_config)``.

    Parameters
    ----------
    func : Callable
        The function to apply to the value (and eventually the flat config)
        to make the new value so that:
        flat_dict[key] = func(flat_dict[key]) or func(flat_dict[key], flat_config)
    processing_type : str, optional
        One of "premerge", "postmerge", "presave", "postload" or "endbuild".
        Timing to apply the value update. In all cases the tag is removed on pre-merge.
        By default "premerge".
    regex : Optional[str]
        The regex to match the key.
    tag_name : Optional[str]
        The tag (without "@") to match the key. The tag is removed on pre-merge.
    order : int, optional
        The pre-merge order. By default 0.0.
    persistent : bool, optional
        If True, the processing will be applied on all keys that have already
        matched the tag before. By nature, using regex make the processing
        always persistent. By default, False.

    Raises
    ------
    ValueError
        If both tag and regex are provided or if none of them are provided.
    ValueError
        If the processing type is not one of "premerge", "postmerge", "presave",
        "postload" or "endbuild".

    Returns
    -------
    processing : Processing
        The processing object with the pre-merge method.


    Examples
    --------
    With the following config and 2 processings:

    .. code_block: yaml

        # config.yaml
        neg_number1: 1
        neg_number2: 1
        neg_number3@add1: 1

    ::

        proc2 = create_processing_value(lambda val: -val, regex="neg_number.*",
            order=0.0)
        proc1 = create_processing_value(lambda val: val + 1, tag_name="add1",
            order=1.0)

    When config.yaml is merged with an other config, it will be considered
    before merging as:

    ::

        {'number1': -1, 'number2': -1, 'number3': 0}

    Using the config as a second argument of the function:

    .. code_block: yaml

        # config.yaml
        param1: 1
        param2@eval: "config.param1 + 1"

    ::

        proc = create_processing_value(
            lambda val, config: eval(val, {'config': config}),
            processing_type='postmerge', tag_name='eval', persistent=False
        )

    After config.yaml is merged with another config, param2 will be evaluated
    as 2 (except if config.param1 has changed with a processing before).
    """
    if tag_name is not None:
        if regex is not None:
            raise ValueError("You must provide a tag or a regex but not both.")
    else:
        if regex is None:
            raise ValueError(
                "You must provide a tag or a regex (to trigger the value update)."
            )
    proc = _ProcessingValue(func, processing_type, regex=regex,
                            tag_name=tag_name, order=order, persistent=persistent)
    return proc


def create_processing_keep_property(
    func: Callable,
    regex: Optional[str] = None,
    tag_name: Optional[str] = None,
    premerge_order: float = 0.0,
    postmerge_order: float = 0.0,
    endbuild_order: float = 0.0,
) -> Processing:
    """Create a processing object that keep a property from a value using tag or regex.

    The pre-merge processing looks for keys that match the tag or the regex, apply
    the function func on the value and store the result (= the "property"):
    ``property = func(flat_dict[key])``.
    The post-merge processing will check that the property is the same as the one
    stored during pre-merge. If not, it will raise a ValueError.

    It also possible to pass the flat config as a second argument of the function
    ``func``. In this case, the function apply
    ``property = func(flat_dict[key], flat_config)``.

    Parameters
    ----------
    func : Callable
        The function to apply to the value (and eventually the flat config)
        to define the property to keep.
        property = func(flat_dict[key]) or func(flat_dict[key], flat_config)
    regex : Optional[str]
        The regex to match the key.
    tag_name : Optional[str]
        The tag (without "@") to match the key. The values are modified when
        triggering the pattern ".*@<tag_name>.*" and the tag is removed from the key.
    premerge_order : float, optional
        The pre-merge order, by default 0.0
    postmerge_order : float, optional
        The post-merge order, by default 0.0
    endbuild_order : float, optional
        The end-build order, by default 0.0

    Raises
    ------
    ValueError
        If both tag and regex are provided or if none of them are provided.

    Returns
    -------
    Processing
        The processing object with the pre-merge and post-merge methods.

    Examples
    --------
    A processing that enforce the types of all the parameters to be constant
    (equal to the type of the first value encountered):

    ::

        create_processing_keep_property(type, regex=".*", premerge_order=15.0,
                                        postmerge_order=15.0, endbuild_order=15.0)

    A processing that protect parameters tagged with @protect from being changed:

    ::

        create_processing_keep_property(lambda x: x, tag_name="protect",
                                        premerge_order=15.0, postmerge_order=15.0)
    """
    if tag_name is not None:
        if regex is not None:
            raise ValueError("You must provide a tag or a regex but not both.")
    else:
        if regex is None:
            raise ValueError(
                "You must provide a tag or a regex (to trigger the value update)."
            )
    processing = _ProcessingKeepProperty(
        func,
        regex=regex,
        tag_name=tag_name,
        premerge_order=premerge_order,
        postmerge_order=postmerge_order,
        endbuild_order=endbuild_order,
    )
    return processing



"""Test the type parser module."""
import re

import pytest
import pytest_check as check

from cliconfig.processing._type_parser import (
    _isinstance,
    _parse_dict,
    _parse_list,
    _parse_optional,
    _parse_type,
    _parse_union,
)


def test_type_parser_isinstance() -> None:
    """Test _parse_type and _isinstance."""
    type_ = _parse_type("None")
    check.equal(type_, (type(None),))
    check.is_true(_isinstance(None, type_))
    check.is_false(_isinstance([None], type_))

    type_ = _parse_type("List[float]")
    check.equal(type_, (("list", (float,)),))

    type_ = _parse_type("list|dict")
    check.equal(type_, (list, dict))
    check.is_true(_isinstance([2.0, True], type_))

    type_ = _parse_type("Dict[str, List[float]]")
    check.equal(type_, (("dict", (str,), (("list", (float,)),)),))
    check.is_true(_isinstance({"a": [1.0, 2.0]}, type_))
    check.is_false(_isinstance({"a": [1.0, 2.0], "b": [1.0, 2]}, type_))

    type_ = _parse_type("Dict[str|bool, int|float]")
    check.equal(type_, (("dict", (str, bool), (int, float)),))

    type_ = _parse_type("Union[Optional[float], Any]")
    check.equal(type_, ((type(None), float), object))
    check.is_true(_isinstance("2", type_))

    type_ = _parse_type(
        "Dict[str, Union[List[None|float], Dict[bool, Optional[list]]]]|List[Any]|"
        "Dict[int, Optional[Dict[str, float]]]|Optional[float]"
    )
    check.equal(
        (
            (
                "dict",
                (str,),
                (
                    ("list", (type(None), float)),
                    ("dict", (bool,), ((type(None), list),)),
                ),
            ),
            ("list", (object,)),
            ("dict", (int,), ((type(None), ("dict", (str,), (float,))),)),
            (type(None), float),
        ),
        type_,
    )
    check.is_true(_isinstance({"a": [None, 1.0], "b": {True: None}}, type_))
    check.is_true(_isinstance([[]], type_))
    check.is_true(_isinstance({1: {"a": 2.0}}, type_))
    check.is_false(_isinstance({"a": [None, 1.0], "b": {False: 1, True: "1"}}, type_))

    # Wrong type description
    with pytest.raises(ValueError, match="Unknown type: 'unknown'"):
        _parse_type("unknown")
    desc = "str, List[float]"
    with pytest.raises(ValueError, match=re.escape(f"Unknown type: '{desc}'")):
        _parse_type(desc)
    desc = "None||float"
    with pytest.raises(ValueError, match=f"Unknown type: '{desc}'"):
        _parse_type(desc)
    desc = (
        "Dict[str, Union[List[None|float], Dict[bool, Optionnal[int]]]]|List[Any]|"
        "Dict[List[int], Optional[Dict[str, float]]]|float"
    )  # Optionnal with 2 n
    with pytest.raises(ValueError, match=f"Unknown type: '{desc}'"):
        _parse_type(desc)

    # Wrong type in isinstance (here a 'dict' tuple has 3 elements instead of 2)
    wrong_type = (
        (
            "dict",
            (str,),
            (
                ("list", (type(None), float)),
                ("dict", (bool,), ((type(None), list),), str),
            ),
        ),
        ("list", (object,)),
        ("dict", (int,), ((type(None), ("dict", (str,), (float,))),)),
        (type(None), float),
    )
    with pytest.raises(ValueError, match="Invalid type for _isinstance:.*"):
        _isinstance({"a": {True: "a"}}, wrong_type)


def test_errors_in_parse() -> None:
    """Test error raised in _parse_X functions."""
    with pytest.raises(ValueError, match="Invalid List type:.*"):
        _parse_list("List[bool, str]")

    with pytest.raises(ValueError, match="Invalid Dict type:.*"):
        _parse_dict("Dict[str, bool, int]]")

    with pytest.raises(ValueError, match="Invalid Union type:.*"):
        _parse_union("Union[str]")

    with pytest.raises(ValueError, match="Invalid Optional type:.*"):
        _parse_optional("Optional[str, int]")



"""Functions to manipulate config as dict with yaml files and CLI."""
import sys
from typing import Any, Dict, List, Optional

from cliconfig.base import Config
from cliconfig.cli_parser import parse_cli
from cliconfig.dict_routines import flatten, show_dict, unflatten
from cliconfig.process_routines import (
    end_build_processing,
    load_processing,
    merge_flat_paths_processing,
    merge_flat_processing,
    save_processing,
)
from cliconfig.processing.base import Processing
from cliconfig.processing.builtin import DefaultProcessings
from cliconfig.tag_routines import clean_all_tags, is_tag_in


def make_config(
    *default_config_paths: str,
    process_list: Optional[List[Processing]] = None,
    add_default_processing: bool = True,
    fallback: str = "",
    no_cli: bool = False,
) -> Config:
    r"""Make a config from default config(s) and CLI argument(s) with processing.

    The function uses the CLI Config routines :func:`.parse_cli` to parse the CLI
    arguments and merge them with :func:`.merge_flat_paths_processing`, applying
    the pre-merge and post-merge processing functions on each merge.

    Parameters
    ----------
    default_config_paths : Tuple[str]
        Paths to default configs. They are merged in order and new keys
        are allowed.
    process_list: Optional[List[Processing]], optional
        The list of processing to apply during each merge. None for empty list.
        By default None.
    add_default_processing : bool, optional
        If add_default_processing is True, the default processings
        (found on :class:`.DefaultProcessings`) are added to the list of
        processings. By default True.
    fallback : str, optional
        Path of the configuration to use if no additional config is provided
        with ``--config``. No fallback config if empty string (default),
        in that case, the config is the default configs plus the CLI arguments.
    no_cli : bool, optional
        If True, the CLI arguments are not parsed and the config is only
        built from the default_config_paths in input and the
        fallback argument is ignored. By default False.

    Raises
    ------
    ValueError
        If additional configs have new keys that are not in default configs.

    Returns
    -------
    config : Config
        The nested built config. Contains the config dict (config.dict) and
        the processing list (config.process_list) which can be used to apply
        further processing routines.

    Note
    ----

        Setting additional arguments from CLI that are not in default configs
        does NOT raise an error but only a warning. This ensures the compatibility
        with other CLI usage (e.g notebook, argparse, etc.)

    Examples
    --------
    ::

        # main.py
        config = make_config('data.yaml', 'model.yaml', 'train.yaml')

    .. code-block:: text

        $ python main.py -- config [bestmodel.yaml,mydata.yaml] \
              --architecture.layers.hidden_dim=64

    """
    # Create the processing list
    process_list_: List[Processing] = [] if process_list is None else process_list
    if add_default_processing:
        process_list_ += DefaultProcessings().list
    config = Config({}, process_list_)
    if no_cli:
        additional_config_paths: List[str] = []
        cli_params_dict: Dict[str, Any] = {}
    else:
        additional_config_paths, cli_params_dict = parse_cli(sys.argv)
        if not additional_config_paths and fallback:
            # Add fallback config
            additional_config_paths = [fallback]
    # Merge default configs and additional configs
    for i, paths in enumerate([default_config_paths, additional_config_paths]):
        # Allow new keys for default configs only
        allow_new_keys = i == 0
        for path in paths:
            config = merge_flat_paths_processing(
                config,
                path,
                allow_new_keys=allow_new_keys,
                preprocess_first=False,  # Already processed
            )

    # Allow new keys for CLI parameters but do not merge them and raise
    # warning.
    cli_params_dict = flatten(cli_params_dict)
    new_keys, keys = [], list(cli_params_dict.keys())
    for key in keys:
        if (
            not is_tag_in(key, "new", full_key=True)
            and clean_all_tags(key) not in config.dict
        ):
            # New key: delete it
            new_keys.append(clean_all_tags(key))
            del cli_params_dict[key]
    if new_keys:
        new_keys_message = "  - " + "\n  - ".join(new_keys)
        print(
            "[CONFIG] Warning: New keys found in CLI parameters "
            f"that will not be merged:\n{new_keys_message}"
        )
    # Merge CLI parameters
    cli_params_config = Config(cli_params_dict, [])
    config = merge_flat_processing(
        config, cli_params_config, allow_new_keys=False, preprocess_first=False
    )
    print(
        f"[CONFIG] Info: Merged {len(default_config_paths)} default config(s), "
        f"{len(additional_config_paths)} additional config(s) and "
        f"{len(cli_params_dict)} CLI parameter(s)."
    )
    # Apply end-build processing
    config = end_build_processing(config)
    # Unflatten the config dict
    config.dict = unflatten(config.dict)
    return config


def load_config(
    path: str,
    default_config_paths: Optional[List[str]] = None,
    process_list: Optional[List[Processing]] = None,
    *,
    add_default_processing: bool = True,
) -> Config:
    """Load config from a file and merge into optional default configs.

    First merge the default configs together (if any), then load the config
    from path, apply the post-load processing, and finally merge the loaded
    config.

    Parameters
    ----------
    path : str
        The path to the file to load the configuration.
    default_config_paths : Optional[List[str]], optional
        Paths to default configs. They are merged in order, new keys are allowed.
        Then, the loaded config is merged into the result. None for no default configs.
        By default None.
    process_list: Optional[List[Processing]]
        The list of processing to apply after loading and for the merges.
        If None, no processing is applied. By default None.
    add_default_processing : bool, optional
        If add_default_processing is True, the default processings
        (found on :class:`.DefaultProcessings`) are added to the list of
        processings. By default True.

    Returns
    -------
    config: Dict[str, Any]
        The nested loaded config. Contains the config dict (config.dict) and
        the processing list (config.process_list) which can be used to apply
        further processing routines.

    Note
    ----

        If default configs are provided, the function does not allow new keys
        for the loaded config. This is for helping the user to see how to
        adapt the config file if the default configs have changed.
    """
    # Crate process_list
    process_list_: List[Processing] = [] if process_list is None else process_list
    if add_default_processing:
        process_list_ += DefaultProcessings().list

    config = Config({}, process_list_)
    if default_config_paths:
        for config_path in default_config_paths:
            config = merge_flat_paths_processing(
                config,
                config_path,
                allow_new_keys=True,
                preprocess_first=False,  # Already processed
            )
    loaded_config = load_processing(path, config.process_list)
    # Update the config list from loaded_config in config
    config.process_list = loaded_config.process_list
    # Merge the loaded config into the config and
    # disallow new keys for loaded config
    # if default configs are provided
    config = merge_flat_processing(
        config,
        loaded_config,
        allow_new_keys=default_config_paths is None,
        preprocess_first=False,
    )
    # Apply end-build processing
    config = end_build_processing(config)
    # Unflatten the config
    config.dict = unflatten(config.dict)
    return config


def save_config(config: Config, path: str) -> None:
    """Save a config and apply pre-save processing before saving.

    Alias for :func:`.save_processing`.

    Parameters
    ----------
    config : Dict[str, Any]
        The config to save.
    path : str
        The path to the yaml file to save the dict.
    """
    save_processing(config, path)


def show_config(config: Config) -> None:
    """Show the config dict in a pretty way.

    The config dict is automatically unflattened before printing.

    Parameters
    ----------
    config : Config
        The config to show.
    """
    print("Config:")
    show_dict(config.dict, start_indent=1)



"""Test base class of configuration."""

import pytest_check as check

from cliconfig.base import Config
from cliconfig.dict_routines import unflatten
from cliconfig.processing.base import Processing


def test_config(process_add1: Processing) -> None:
    """Test base class of configuration."""
    config_dict = {"a.b": 1, "b": 2, "c.d.e": 3, "c.d.f": [2, 3]}
    config_dict = unflatten(config_dict)
    config = Config(config_dict, [process_add1])
    check.equal(config.dict, config_dict)
    check.equal(config.process_list, [process_add1])
    check.equal(repr(config), f"Config({config_dict}, ['ProcessAdd1'])")
    config_dict2 = {"c.d.e": 3, "a.b": 1, "b": 2, "c.d.f": [2, 3]}
    config_dict2 = unflatten(config_dict2)
    config2 = Config(config_dict2, [process_add1])
    check.equal(config, config2)
    config3 = Config(config_dict2, [process_add1, process_add1])
    check.not_equal(config, config3)
    config4 = Config(config_dict2, [])
    check.not_equal(config, config4)
    # Test get/set attribute
    config = Config(config_dict, [process_add1])
    check.equal(config.a.b, 1)
    check.equal(config.c.d.f, [2, 3])
    check.equal(config.c.d, Config({"e": 3, "f": [2, 3]}, [process_add1]))
    config.a.b = 2
    check.equal(config.a.b, 2)
    config.c.d.f = [3, 4]
    check.equal(config.c.d.f, [3, 4])
    config.c.d.f.append(5)
    check.equal(config.c.d.f, [3, 4, 5])
    # Test delete attribute
    del config.a.b
    check.equal(config.dict, {"a": {}, "b": 2, "c": {"d": {"e": 3, "f": [3, 4, 5]}}})
    del config.c.d
    check.equal(config.dict, {"a": {}, "b": 2, "c": {}})
    # Should no raise error
    del config.dict
    del config.process_list



"""Test helpers to create processing functions."""
import re

import pytest
import pytest_check as check

from cliconfig.base import Config
from cliconfig.processing.create import (
    _is_matched,
    create_processing_keep_property,
    create_processing_value,
)


def test_is_matched() -> None:
    """Test _is_matched."""
    check.is_true(_is_matched("foo.bar.test", regex=".*st", tag_name=None))
    check.is_false(_is_matched("test.bar.foo", regex=".*st", tag_name=None))
    check.is_true(_is_matched("foo.bar.test@tag", regex=None, tag_name="tag"))
    check.is_false(_is_matched("foo.bar@tag.test", regex=None, tag_name="tag"))
    check.is_false(_is_matched("foo.bar.test@tag2", regex=None, tag_name="tag"))
    with pytest.raises(
        ValueError, match="Either regex or tag_name must be defined but not both."
    ):
        _is_matched("foo.bar", ".*", "tag")
    with pytest.raises(
        ValueError, match="Either regex or tag_name must be defined."
    ):
        _is_matched("foo.bar", None, None)


def test_create_processing_value() -> None:
    """Test create_processing_value."""
    # Persistent, premerge
    proc1 = create_processing_value(
        lambda x: x + 1, tag_name="add1", order=1.0, persistent=True
    )
    proc2 = create_processing_value(
        lambda x: -x, regex="neg_number.*", order=0.0, persistent=True
    )
    in_dict = {"neg_number1": 1, "neg_number2": 1, "neg_number3@add1": 1}
    config = Config(in_dict, [proc1, proc2])
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(config.dict, {"neg_number1": -1, "neg_number2": -1, "neg_number3": 0})
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    config = config.process_list[1].postmerge(config)
    config = config.process_list[0].postmerge(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    config = config.process_list[1].endbuild(config)
    config = config.process_list[0].endbuild(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    config = config.process_list[1].presave(config)
    config = config.process_list[0].presave(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    config = config.process_list[1].postload(config)
    config = config.process_list[0].postload(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    # Non persistent, postmerge
    proc2 = create_processing_value(
        lambda x: -x, "postmerge", regex="neg_number.*", order=0.0, persistent=False
    )
    proc1 = create_processing_value(
        lambda x: x + 1, "postmerge", tag_name="add1", order=1.0, persistent=False
    )
    in_dict = {"neg_number1": 1, "neg_number2": 1, "neg_number3@add1": 1}
    config = Config(in_dict, [proc1, proc2])
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    config = config.process_list[1].postmerge(config)
    config = config.process_list[0].postmerge(config)
    check.equal(config.dict, {"neg_number1": -1, "neg_number2": -1, "neg_number3": 0})
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    config = config.process_list[1].postmerge(config)
    config = config.process_list[0].postmerge(config)
    check.equal(config.dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 0})
    # No signature, endbuild, presave, postload
    proc1 = create_processing_value(type, "endbuild", tag_name="get_type")
    proc2 = create_processing_value(str, "presave", tag_name="to_str")
    proc3 = create_processing_value(
        lambda *args: 0,  # noqa
        "postload",
        tag_name="zero"
    )
    in_dict2 = {
        "type1@get_type@to_str@zero": 1,
        "type2@get_type@to_str@zero": "2",
        "type3@get_type@to_str@zero": True
    }
    config = Config(in_dict2, [proc1, proc2, proc3])
    config = config.process_list[0].premerge(config)
    config = config.process_list[1].premerge(config)
    config = config.process_list[2].premerge(config)
    check.equal(config.dict, {"type1": 1, "type2": "2", "type3": True})
    config = config.process_list[0].endbuild(config)
    check.equal(config.dict, {"type1": int, "type2": str, "type3": bool})
    config = config.process_list[1].presave(config)
    check.equal(config.dict, {
        "type1": "<class 'int'>",
        "type2": "<class 'str'>",
        "type3": "<class 'bool'>"
    })
    config = config.process_list[2].postload(config)
    check.equal(config.dict, {"type1": 0, "type2": 0, "type3": 0})

    # Config in input
    proc = create_processing_value(
        lambda x, flat_config: eval(  # pylint: disable=eval-used
            x, {"config": flat_config}
        ),
        tag_name="eval",
        persistent=False,
    )
    eval_dict = {"param1": 1, "param2@eval": "config.param1 + 1"}
    config = Config(eval_dict, [proc])
    config = config.process_list[0].premerge(config)
    check.equal(config.dict, {"param1": 1, "param2": 2})

    # Failing cases
    with pytest.raises(
        ValueError, match="Processing type 'UNKNOWN' not recognized.*"
    ):
        create_processing_value(lambda x: x + 1, "UNKNOWN", tag_name="tag")
    with pytest.raises(
        ValueError, match="You must provide a tag or a regex but not both."
    ):
        create_processing_value(lambda x: x + 1, tag_name="add1", regex="neg_number.*")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "You must provide a tag or a regex (to trigger the value update)."
        ),
    ):
        create_processing_value(lambda x: x + 1, order=0.0)


def test_create_processing_keep_property() -> None:
    """Test create_processing_keep_property."""
    proc1 = create_processing_keep_property(type, regex=".*")
    proc2 = create_processing_keep_property(lambda x: x, tag_name="protect")
    proc3 = create_processing_keep_property(
        lambda x, config: len(x + config.dict["list"]),
        tag_name="len",
    )
    in_dict = {"str": "foo", "int": 2, "list@protect": [1, 2, 3], "list2@len": [1, 2]}
    config = Config(in_dict, [proc1, proc2, proc3])
    config = config.process_list[0].premerge(config)
    config = config.process_list[1].premerge(config)
    config = config.process_list[2].premerge(config)
    check.equal(
        config.dict,
        {"str": "foo", "int": 2, "list": [1, 2, 3], "list2": [1, 2]}
    )
    config.dict = {"str": "foo", "int": -1, "list": [1, 2, 3], "list2": [4, 5]}
    config = config.process_list[0].postmerge(config)
    config = config.process_list[1].postmerge(config)
    config = config.process_list[2].postmerge(config)
    config.dict = {"str": "foo2", "int": 6, "list": [1, 2, 3], "list2": [8, 9]}
    config = config.process_list[0].endbuild(config)
    config = config.process_list[1].endbuild(config)
    config = config.process_list[2].endbuild(config)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Property of key str has changed from <class 'str'> to <class 'int'> "
            "while it is protected by a keep-property processing (problem found on "
            "end-build)."
        ),
    ):
        config.process_list[0].endbuild(Config({"str": 0}, []))
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Property of key list has changed from [1, 2, 3] to [1, 2] while "
            "it is protected by a keep-property processing (problem found on "
            "post-merge)."
        ),
    ):
        config.process_list[1].postmerge(Config({"list": [1, 2]}, []))
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Property of key list2 has changed from 5 to 3 while "
            "it is protected by a keep-property processing (problem found on "
            "post-merge)."
        ),
    ):
        config.process_list[2].postmerge(Config({"list": [1, 2], "list2": [3]}, []))
    # Should not raise on pre-merge:
    config.process_list[0].premerge(Config({"str": 0}, []))
    config.process_list[1].premerge(Config({"list": [1, 2]}, []))
    config.process_list[2].premerge(Config({"list": [1, 2], "list2": []}, []))

    # Failing cases
    with pytest.raises(
        ValueError, match="You must provide a tag or a regex but not both."
    ):
        create_processing_keep_property(lambda x: x, tag_name="protect", regex=".*")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "You must provide a tag or a regex (to trigger the value update)."
        ),
    ):
        create_processing_keep_property(lambda x: x, premerge_order=0.0)
